from typing import List, Union, Dict, Optional, Callable

import numpy as np
import matplotlib.pyplot as plt

from quickstats import GeneralEnum

class StatMeasure(GeneralEnum):
    MEAN   = (0, np.mean)
    STD    = (1, np.std)
    MIN    = (2, np.min)
    MAX    = (3, np.max)
    MEDIAN = (4, np.median)
    
    def __new__(cls, value:int, operator:Callable):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.operator = operator
        return obj           

class StatPlotConfig:
    
    @property
    def stat_measures(self):
        return self._stat_measures
    
    @stat_measures.setter
    def stat_measures(self, values):
        parsed = []
        for value in values:
            parsed.append(StatMeasure.parse(value))
        self._stat_measures = parsed            
    
    def __init__(self, stat_measures:List[Union[StatMeasure, str]],
                 axis_method:str, options:Dict,
                 handle_options:Optional[Dict]=None,
                 handle_return_method:Optional[Callable]=None):
        self.stat_measures  = stat_measures
        self.axis_method    = axis_method
        self.options        = options
        if handle_options is None:
            self.handle_options = self.get_default_handle_options()
        else:
            self.handle_options = handle_options
        self.quantities     = {}
        self.handle_return_method = handle_return_method
        
    def set_data(self, x:np.ndarray):
        quantities = {}
        for stat_measure in self.stat_measures:
            quantity = stat_measure.operator(x)
            name = stat_measure.name.lower()
            quantities[name] = quantity
        self.quantities = quantities
    
    def get_default_handle_options(self):
        return {}
    
    def apply(self, ax, main_handle:Optional=None):
        if not hasattr(ax, self.axis_method):
            raise RuntimeError(f"matplotlib.axes.Axes has no method named {self.axis_method}")
        if not self.quantities:
            raise RuntimeError(f"stat data not initialized")
        method = getattr(ax, self.axis_method)
        resolved_options = {}
        for name in self.options:
            if callable(self.options[name]):
                resolved_options[name] = self.options[name](self.quantities)
            else:
                resolved_options[name] = self.options[name]
        if (main_handle is not None) and (self.handle_options is not None):
            for name in self.handle_options:
                if name in resolved_options:
                    continue
                resolved_options[name] = self.handle_options[name](main_handle)
        result = method(**resolved_options)
        if self.handle_return_method is not None:
            return self.handle_return_method(result)
        return result

class HandleMatchConfig(StatPlotConfig):
    def get_default_handle_options(self):
        handle_options = {
            "color": lambda handle: handle.get_color()
        }
        return handle_options
    
class AverageLineH(HandleMatchConfig):
    def __init__(self, **styles):
        options = {
            "y":  lambda x: x["mean"],
            **styles
        }
        super().__init__(["mean"], "axhline", options=options)
        
class AverageLineV(HandleMatchConfig):
    def __init__(self, **styles):
        options = {
            "x":  lambda x: x["mean"],
            **styles
        }
        super().__init__(["mean"], "axvline", options=options)        
        
class StdBandH(HandleMatchConfig):
    def __init__(self, **styles):
        options = {
            "ymin": lambda x: x["mean"] - x["std"],
            "ymax": lambda x: x["mean"] + x["std"],
            **styles
        }
        super().__init__(["mean", "std"], "axhspan", options=options)
        
class StdBandV(HandleMatchConfig):
    def __init__(self, **styles):
        options = {
            "xmin": lambda x: x["mean"] - x["std"],
            "xmax": lambda x: x["mean"] + x["std"],
            **styles
        }
        super().__init__(["mean", "std"], "axvspan", options=options)        


from typing import Optional, List, Dict
import re

from .rooproc_helper_action import RooProcHelperAction

class RooProcGlobalVariables(RooProcHelperAction):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def commarepl(matchobj):
        return matchobj.group(0).replace(",", "::COMMA::")        
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        main_text = re.sub(r"\s*", "", main_text)
        main_text = re.sub(r'\[(.*?)\]', cls.commarepl, main_text)
        tokens = [token.replace("::COMMA::", ",") for token in main_text.split(",")]
        global_variables = {}
        for token in tokens:
            result = re.match("^(\w+)=(.*)", token)
            if not result:
                raise RuntimeError(f"invalid expression {token}")
            global_variables[result[1]] = result[2]
        return cls(**global_variables)
    
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        processor.global_variables.update(params)
        return processor


from .rooproc_define import RooProcDefine

class RooProcSafeDefine(RooProcDefine):
        
    def _execute(self, rdf:"ROOT.RDataFrame", **params):
        name = params['name']
        expression = params['expression']
        all_column_names = [str(i) for i in rdf.GetColumnNames()]
        # already defined, skipping
        if name in all_column_names:
            return rdf
        rdf_next = rdf.Define(name, expression)
        return rdf_next


from typing import Optional, Union, Dict, List
import os
import json
import glob

from quickstats import semistaticmethod

class ConfigParser:
    @staticmethod
    def validate_file(filename:str):
        if not os.path.exists(filename):
            raise FileNotFoundError(f"configuration file \"{filename}\" does not exist")
    
    


from typing import Optional, List

from .decorators import semistaticmethod

class TVirtualNode:
    
    def __init__(self, name:Optional[str]=None, level:Optional[int]=0,
                 parent:Optional["TVirtualNode"]=None,
                 **data):
        self.name     = name
        self.level    = level
        self.parent   = parent
        self.prev_sibling = None
        self.next_sibling = None
        self.children = []
        self.data     = {**data}
        
    def get_number_of_children(self):
        return len(self.children)
    
    def get_data(self, key:str):
        if key not in self.data:
            raise RuntimeError(f'missing data attribute "{key}"')
        return self.data[key]
    
    @property
    def has_child(self):
        return self.get_number_of_children() > 0
    
    @property
    def first_child(self):
        if self.has_child:
            return self.children[0]
        return None
    
    @property
    def last_child(self):
        if self.has_child:
            return self.children[-1]
        return None
        
class TVirtualTree:
    
    NodeClass = TVirtualNode
    
    def __init__(self):
        self.root_node    = self.NodeClass()
        self.current_node = self.root_node
    
    def add_child(self, name:str, **data):
        child_node = self.NodeClass(name, level=self.current_node.level + 1,
                                    parent=self.current_node, **data)
        if self.current_node.has_child:
            last_child = self.current_node.last_child
            child_node.prev_sibling = last_child
            last_child.next_sibling = child_node
        self.current_node.children.append(child_node)
        return child_node
    
    @semistaticmethod
    def _to_diagram(self, source, indent:str=' '*4, level:int=0):
        if not isinstance(source, self.NodeClass):
            raise TypeError("source must be ActionNode, not %s" % (source.__class__.__name__,))
        if source.name is not None:
            string = f"{indent * level}- {source.name}\n"
        else:
            string = ""
            level  = - 1
        for child_node in source.children:
            string += self._to_diagram(child_node, indent=indent, level=level + 1)
        return string
    
    def __str__(self):
        return self._to_diagram(self.root_node)
    
    def reset(self):
        self.current_node = self.root_node
    
    def get_next(self, consider_child:bool=True):
        node = self.current_node
        if consider_child and node.has_child:
            node = node.first_child
        else:
            while True:
                if node.next_sibling is not None:
                    node = node.next_sibling
                    break
                elif node.parent is not None:
                    node = node.parent
                else:
                    node = None
                    break
        self.current_node = node
        return node


from typing import Optional
import re

from .rooproc_hybrid_action import RooProcHybridAction

class RooProcStat(RooProcHybridAction):
    def __init__(self, ext_var_name:str, column_name:str):
        super().__init__(ext_var_name=ext_var_name,
                         column_name=column_name)
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        name_literals = re.findall(r"@{([^{}]+)}", main_text)
        if len(name_literals) == 1:
            ext_var_name = name_literals[0]
            column_name = main_text.replace("@{" + ext_var_name + "}", "").strip()
        else:
            raise RuntimeError(f"unspecified external variable name (format:@{{ext_var_name}}): {main_text}")
        return cls(ext_var_name=ext_var_name, column_name=column_name)
    
    def _get_func(self, rdf:"ROOT.RDataFrame"):
        raise NotImplementedError
        
    def _execute(self, rdf:"ROOT.RDataFrame", processor:"quickstats.RooProcessor", **params):
        ext_var_name = params['ext_var_name']
        column_name = params['column_name']
        processor.external_variables[ext_var_name] = self._get_func(rdf)(column_name)
        return rdf, processor


import os
import re
import json
import time
from typing import Optional, Union, List, Dict


from quickstats.parsers import ConfigParser
    
    
class WSModifierConfig(ConfigParser):
    
    kConfigFormat = {
        'input_file': {
            'required': True,
            'type': str
        },
        'output_file': {
            'required': False,
            'description': 'Path to the location of training scripts ' +\
                           ' (or the directory containing the training scripts)',
            'required': True,
            'type': str
        },
        'model_config': {
            'abbr': 'm',
            'description': 'Name of the model configuration to use',
            'required': True,
            'type': dict
        },
        'search_space': {
            'abbr': 's',
            'description': 'Name of the search space configuration to use',
            'required': True,
            'type':dict
        },
        'hpo_config': {
            'abbr': 'o',
            'description': 'Name of the hpo configuration to use',
            'required': True,
            'type':dict
        },
        'grid_config': {
            'abbr': 'g',
            'description': 'Name of the grid configuration to use',
            'required': True,
            'type':dict
        }
    }
    
    def __init__(self, source:Union[str, Dict]):
        pass
    
    def parse_config(self, source:Union[str, Dict]):
        pass
    
    @classmethod
    def read_json(cls, filename:str):
        self.validate_file(filename)
    
    @classmethod
    def read_xml(cls, filename:str):
        self.validate_file(filename)
    
    def to_json(self, path:str):
        pass
    
    def to_xml(self, path:str):
        pas



from typing import List, Optional, Union, Dict, Callable, Tuple

import ROOT

from quickstats import AbstractObject, semistaticmethod

class PdfFitTool(AbstractObject):
    
    def __init__(self, pdf:"ROOT.RooAbsPdf", data:"ROOT.RooAbsData",
                 verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)
        
        self.pdf        = pdf
        self.data       = data
        
    @staticmethod
    def is_fit_success(fit_result:"ROOT.RooFitResult"):
        status   = fit_result.status()
        cov_qual = fit_result.covQual()
        return (status == 0) and (cov_qual in [-1, 3])
    
    @staticmethod
    def _set_pdf_param_values(pdf:ROOT.RooAbsPdf, observable:ROOT.RooRealVar, param_values:Dict):
        params = pdf.getParameters(observable)
        for param in params:
            param_name = param.GetName()
            if param_name not in param_values:
                raise RuntimeError(f"missing value for the parameter: {param_name}")
            param_value = param_values[param_name]
            param.setVal(param_value)
            
    @semistaticmethod
    def _get_fit_stats(self, model:"ROOT.RooAbsPdf", data:"ROOT.RooAbsData",
                       n_bins:Optional[int]=None, n_float_params:int=0):
        observable = self.get_observable(data)
        if n_bins is None:
            n_bins = observable.numBins()
        # +1 is there to account for the normalization that is done internally in RootFit
        ndf = n_bins - (n_float_params + 1)
        frame = observable.frame()
        data.plotOn(frame)
        model.plotOn(frame)
        chi2_reduced = frame.chiSquare(n_float_params)
        chi2 = chi2_reduced * ndf
        pvalue = ROOT.TMath.Prob(chi2, ndf)
        fit_stats = {
            'n_bins': n_bins,
            'n_float_params': n_float_params,
            'ndf': ndf,
            'chi2/ndf': chi2_reduced,
            'chi2': chi2,
            'pvalue': pvalue
        }
        return fit_stats
    
    def get_fit_stats(self, n_bins:Optional[int]=None, n_float_params:int=0):
        """
        Parameters
            n_bins: (optional) int
                Number of bins used for chi2 calculation. If not specified, the number of bins of the 
                observable is used.
            n_float_params: int, default = 0
                Number of floating parameters in the fit. This decreases the number of degrees of freedom
                used in chi2 calculation.
        """
        return self._get_fit_stats(self.pdf, self.data, n_bins=n_bins, n_float_params=n_float_params)
    
    @semistaticmethod
    def print_fit_stats(self, fit_stats:Dict):
        self.stdout.info(f"INFO: chi^2/ndf = {fit_stats['chi2/ndf']}, "
                         f"Number of Floating Parameters + Normalization = {fit_stats['n_float_params'] + 1}, "
                         f"Number of bins = {fit_stats['n_bins']}, "
                         f"ndf = {fit_stats['ndf']}, "
                         f"chi^2 = {fit_stats['chi2']}, "
                         f"p_value = {fit_stats['pvalue']}")
    
    @staticmethod
    def get_observable(data:"ROOT.RooAbsData"):
        parameters = data.get()
        if len(parameters) > 1:
            raise RuntimeError("only single-observable fit is allowed")
        observable = parameters.first()
        return observable
        
    def mle_fit(self, minos:bool=False, hesse:bool=True, sumw2:bool=True, strategy:int=1,
                min_fit:int=2, max_fit:int=3, range_expand_rate:Optional[int]=None, print_level:int=-1):
        
        observable = self.get_observable(self.data)
        vmin = observable.getMin()
        vmax = observable.getMax()
        observable.setRange("fitRange", vmin, vmax)
        
        model_name = self.pdf.GetName()
        data_name  = self.data.GetName()
        obs_name   = observable.GetName()
        
        self.stdout.info(f"INFO: Begin model fitting...")
        self.stdout.info(f"      Model : ".rjust(20) + f"{model_name}")
        self.stdout.info(f"    Dataset : ".rjust(20) + f"{data_name}")
        self.stdout.info(f" Observable : ".rjust(20) + f"{obs_name}")
        
        fit_args = [ROOT.RooFit.Range("fitRange"), ROOT.RooFit.PrintLevel(print_level),
                    ROOT.RooFit.Minos(minos), ROOT.RooFit.Hesse(hesse),
                    ROOT.RooFit.Save(), ROOT.RooFit.SumW2Error(sumw2),
                    ROOT.RooFit.Strategy(strategy)]

        status_label = {
            True  : 'SUCCESS',
            False : 'FAIL'
        }

        for i in range(1, max_fit + 1):
            fit_result = self.pdf.fitTo(self.data, *fit_args)
            is_success = self.is_fit_success(fit_result)
            self.stdout.info(f" Fit iteration {i} : ".rjust(20) + f"{status_label[is_success]}")
            if i >= min_fit:
                if is_success:
                    return fit_result
                elif range_expand_rate is not None:
                    new_vmin = observable.getRange("fitRange").first - range_expand_rate
                    new_vmax = observable.getRange("fitRange").second + range_expand_rate
                    self.stdout.info(f"INFO: Fit failed to converge, refitting with "
                                     f"expanded fit range [{new_vmin}, {new_vmax}]")
                    observable.setRange("fitRange", new_vmin, new_vmax)
        return fit_result


from typing import Dict, Union, List, Optional

from quickstats.maths.numerics import get_proper_ranges

class RooRealVar:
    
    @property
    def range(self):
        return self._range
    
    @range.setter
    def range(self, value:Optional[List[float]]=None):
        self._range = self._parse_value_range(value, self.value)
   
    @property
    def named_ranges(self):
        return self._named_ranges
    
    @named_ranges.setter
    def named_ranges(self, value:Optional[Dict[str, List[float]]]=None):
        self._named_ranges = self._parse_named_ranges(value)
        
    def __init__(self, obj:Optional[Union[Dict, "ROOT.RooRealVar"]]=None):
        if obj is not None:
            self.parse(obj)
        else:
            self.name   = None
            self.title  = None
            self.value  = None
            self.n_bins = None
            self.unit   = None
            self.range  = None
            self.named_ranges = None
        
    def parse(self, obj:Union[str, Dict, "ROOT.RooRealVar"]):
        if isinstance(obj, str):
            self.parse({"name": obj})
        elif isinstance(obj, dict):
            self.name = obj["name"]
            self.title = obj.get("title", None)
            self.value = obj.get("value", None)
            self.range = obj.get("range", None)
            self.named_ranges = obj.get("named_ranges", None)
            self.n_bins = obj.get("n_bins", None)
            self.unit = obj.get("unit", None)
        else:
            import ROOT
            if not isinstance(obj, ROOT.RooRealVar):
                raise ValueError("object must be an instance of ROOT.RooRealVar")
            self.name   = str(obj.GetName())
            self.title  = str(obj.getTitle())
            self.value  = obj.getVal()
            self.n_bins = obj.getBins()
            self.unit   = obj.getUnit()
            _range = obj.getRange()
            self.range = [_range[0], _range[1]]
            named_ranges = {}
            for name in obj.getBinningNames():
                if str(name) == "":
                    continue
                _range = obj.getRange(name)
                named_ranges[name] = [_range[0], _range[1]]
            self.named_ranges = named_ranges
        
    @staticmethod
    def _parse_named_ranges(named_ranges:Optional[Dict[str, List[float]]]=None):
        if named_ranges:
            ranges = get_proper_ranges(list(named_ranges.values()), no_overlap=False)
            result = {k:v for k, v in zip(named_ranges.keys(), ranges)}
            return result
        return None
    
    @staticmethod
    def _parse_value_range(value_range:Optional[List[float]]=None,
                           nominal_value:Optional[float]=None):
        if value_range is not None:
            result = get_proper_ranges(value_range, nominal_value)
            if result.shape != (1, 2):
                raise ValueError("value range must be list of size 2")
            return result[0]
        return None
                                
    @classmethod
    def create(cls, name:str, title:Optional[str]=None,
               value:Optional[float]=None,
               range:Optional[List[float]]=None,
               named_ranges:Optional[Dict[str, List[float]]]=None,
               n_bins:Optional[int]=None, unit:Optional[str]=None):
        instance = cls()
        kwargs = {
                    "name" : name,
                   "title" : title,
                   "value" : value,
                   "range" : range,
            "named_ranges" : named_ranges,
                  "n_bins" : n_bins,
                    "unit" : unit
        }
        instance.parse(kwargs)
        return instance
        
    def new(self):
            
        if self.name is None:
            raise RuntimeError("object not initialized")
            
        import ROOT
        
        if self.title is not None:
            title = self.title
        else:
            title = self.name
        
        if (self.value is not None) and (self.range is not None):
            variable = ROOT.RooRealVar(self.name, title, self.value,
                                       self.range[0], self.range[1])
        elif (self.value is None) and (self.range is not None):
            variable = ROOT.RooRealVar(self.name, title,
                                       self.range[0], self.range[1])
        elif (self.value is not None) and (self.range is None):
            variable = ROOT.RooRealVar(self.name, title, self.value)
        else:
            variable = ROOT.RooRealVar(self.name, title, 0.)
            
        if self.named_ranges is not None:
            for name, _range in self.named_ranges.items():
                variable.setRange(name, _range[0], _range[1])            
        
        if self.n_bins is not None:
            variable.setBins(self.n_bins)
            
        if self.unit is not None:
            variable.setUnit(self.unit)
            
        return variable


from typing import Optional, Union, Dict, List

import numpy as np
import matplotlib.patches as patches

from quickstats.plots import AbstractPlot
from quickstats.plots.template import create_transform
from quickstats.utils.common_utils import combine_dict

class UpperLimit3DPlot(AbstractPlot):
    
    STYLES = {
        'axis':{
            'tick_bothsides': True
        },
        'errorbar': {
            "linewidth": 1,
            "markersize": 5,
            "marker": 'o',
        }        
    }
    
    COLOR_PALLETE = {
        '2sigma': '#FDC536',
        '1sigma': '#4AD9D9',
        'expected': 'k',
        'excluded': 'w',
        'observed': 'k'
    }
    
    COLOR_PALLETE_SEC = {
        '2sigma': '#FDC536',
        '1sigma': '#4AD9D9',
        'expected': 'r',
        'observed': 'r'
    }
    
    LABELS = {
        '2sigma': 'Expected limit $\pm 2\sigma$',
        '1sigma': 'Expected limit $\pm 1\sigma$',
        'expected': 'Expected limit (95% CL)',
        'observed': 'Observed limit (95% CL)'
    }
    
    LABELS_SEC = {
        '2sigma': 'Expected limit $\pm 2\sigma$',
        '1sigma': 'Expected limit $\pm 1\sigma$',
        'expected': 'Expected limit (95% CL)',
        'observed': 'Observed limit (95% CL)'
    }

    CONFIG = {
        'primary_hatch'  : '\\\\\\',
        'secondary_hatch': '///',
        'primary_alpha'  : 0.9,
        'secondary_alpha': 0.8,
        'curve_line_styles': {
            'color': 'darkred' 
        },
        'curve_fill_styles':{
            'color': 'hh:darkpink'
        },
        'highlight_styles': {
            'linewidth' : 0,
            'marker' : '*',
            'markersize' : 20,
            'color' : '#E9F1DF',
            'markeredgecolor' : 'black'
        }
    }
    
    def __init__(self, data, data_sec=None,
                 num_grid_points=None,
                 color_pallete:Optional[Dict]=None,
                 color_pallete_sec:Optional[Dict]=None,
                 labels:Optional[Dict]=None,
                 labels_sec:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Union[Dict, str]]='default',
                 config:Optional[Dict]=None):
        super().__init__(color_pallete=color_pallete,
                         color_pallete_sec=color_pallete_sec,
                         styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
        self.data     = data
        # secondary data
        self.data_sec = data_sec
        
        self.labels = combine_dict(self.LABELS, labels)
        self.labels_sec = combine_dict(self.LABELS_SEC, labels_sec)
        
        self.num_grid_points = num_grid_points
        self.highlight_data = None
        
    def get_default_legend_order(self):
        return ['observed', 'expected', '1sigma', '2sigma', 'curve', 'highlight']
    
    def add_highlight(self, x:float, y:float, label:str="SM prediction",
                      styles:Optional[Dict]=None):
        highlight_data = {
            'x'     : x,
            'y'     : y,
            'label' : label,
            'styles': styles
        }
        self.highlight_data = highlight_data        
    
    def draw_highlight(self, ax, data):
        styles = data['styles']
        if styles is None:
            styles = self.config['highlight_styles']
        handle = ax.plot(data['x'], data['y'], label=data['label'], **styles)
        self.update_legend_handles({'highlight': handle[0]})

    @staticmethod
    def get_grid(X_range, Y_range, num_grid_points):
        X_points = np.linspace(*X_range, num_grid_points)
        Y_points = np.linspace(*Y_range, num_grid_points)
        X_grid, Y_grid = np.meshgrid(X_points, Y_points)
        return X_grid, Y_grid

    def return_points(self, paths):
        ''' Get x-y coordinates of contour/path in order to fill color for bands '''
        uout, inn = 0, 1
        segment1_start = paths[0].vertices[0]
        segment1_end = paths[0].vertices[-1]
        try:
            if len(paths) == 2 and (segment1_start != segment1_end).all():
                segment2_start = paths[1].vertices[0]
                segment2_end = paths[1].vertices[-1]
                if np.square(segment2_start-segment2_end) > np.square(segment1_start-segment1_end):
                    out, inn = 1, 0
                return (paths[out].vertices, paths[inn].vertices)
            else:
                return (np.concatenate([i.vertices for i in paths]),)
        except:
            return (np.concatenate([i.vertices for i in paths]),)

    def plot_bands(self, ax, cp, labels, cp_obs=None):
        # loop over contour curves
        # assumption: 0,4 = 2sigma outer/inner boundaries; 1,3 = 1sigma outer/inner boundaries; 2 = expectation
        inner2 = self.return_points(cp.collections[0].get_paths())
        inner1 = self.return_points(cp.collections[1].get_paths())
        center = self.return_points(cp.collections[2].get_paths())
        outer1 = self.return_points(cp.collections[3].get_paths())
        outer2 = self.return_points(cp.collections[4].get_paths())
        if cp_obs:
            observ = self.return_points(cp_obs.collections[0].get_paths())

        # fill color from outside to inside
        handle_2sigma = ax.fill(outer2[0][:,0], outer2[0][:,1], self.color_pallete['2sigma'], label=labels['2sigma'], zorder=1)
        handle_1sigma = ax.fill(outer1[0][:,0], outer1[0][:,1], self.color_pallete['1sigma'], label=labels['1sigma'], zorder=1)
        handle_expected = ax.plot(center[0][:,0], center[0][:,1], self.color_pallete['expected'], ls='--', label=labels['expected'], zorder=2)
        ax.fill(inner1[0][:,0], inner1[0][:,1], self.color_pallete['2sigma'], zorder=1)
        ax.fill(inner2[0][:,0], inner2[0][:,1], self.color_pallete['excluded'], zorder=1)
        if len(outer2) > 1:
            handle_2sigma = ax.fill(outer2[1][:,0], outer2[1][:,1], self.color_pallete['excluded'], zorder=1)
        if len(outer1) + len(center) + len(inner1) + len(inner2) > 4 :
            print('WARN', len(outer1) , len(center) , len(inner1) , len(inner2))

        if cp_obs:
            handle_observed = ax.plot(observ[0][:,0], observ[0][:,1], self.color_pallete['observed'], ls='-', label=labels['observed'], zorder=2)

        self.update_legend_handles({'2sigma': handle_2sigma[0], '1sigma': handle_1sigma[0], 'expected': handle_expected[0]})
        if cp_obs:
            self.update_legend_handles({'observed': handle_observed[0]})

    def draw_single_data(self, ax, df, x='klambda', y='k2v', scale_factor=None,
                         theory_grid=None,
                         color_pallete:Optional[Dict]=None,
                         labels:Optional[Dict]=None,
                         draw_observed:bool=False,
                         observed_marker:Optional[str]='o', 
                         alpha:float=1.,
                         sec:bool=False):
        
        assert(theory_grid.shape == (self.num_grid_points, self.num_grid_points)), '`theory_grid` in draw() fails to have shape of (`self.num_grid_points`, `self.num_grid_points`)'
        assert(not df.isnull().values.any()), 'input dataframe contains NAN'
        if color_pallete is None:
            color_pallete = self.color_pallete
        if labels is None:
            labels = self.labels
        if scale_factor is None:
            scale_factor = 1.0

        # Create 2D grid of values for a range of two coupling constant variations (e.g. a grid of k2v and kl points) by interpolating to the grid
        X_range = df[x].min(), df[x].max()
        Y_range = df[y].min(), df[y].max()
        X_grid, Y_grid = UpperLimit3DPlot.get_grid(X_range, Y_range, self.num_grid_points)

        expectation_point_list = {}
        expectation_point_list['points'] = (df[x], df[y])
        expectation_point_list['exp'] = df['0'] * scale_factor
        expectation_point_list['1sigma'] = df['1'] * scale_factor
        expectation_point_list['-1sigma'] = df['-1'] * scale_factor
        expectation_point_list['2sigma'] = df['2'] * scale_factor
        expectation_point_list['-2sigma'] = df['-2'] * scale_factor
        if draw_observed:
            expectation_point_list['obs'] = df['obs'] * scale_factor
        
        exp_grids = {}
        for i in expectation_point_list.keys():
            if i == 'points':
                continue
            from scipy import interpolate
            exp_grids[i] = interpolate.griddata(expectation_point_list['points'], 
                                                expectation_point_list[i], np.stack((X_grid, Y_grid), axis=2))
        
        # Create a 2D grid of Z-values for contour plot, arbitarily assign values of:
        #   0.5 within +1 sigma, 1.5 between +1 and +2 sigma, 2.5 outside of 2 sigma;
        #   -0.5 within -1 sigma, -1.5 between -1 and -2 sigma, -2.5 outside of -2 sigma
        Z_grid = np.zeros(theory_grid.shape)
        Z_grid[theory_grid - exp_grids['exp'] >= 0] = 0.5
        Z_grid[theory_grid - exp_grids['exp'] < 0] = -0.5
        Z_grid[theory_grid - exp_grids['1sigma'] > 0] = 1.5
        Z_grid[theory_grid - exp_grids['2sigma'] > 0] = 2.5
        Z_grid[theory_grid - exp_grids['-1sigma'] < 0] = -1.5
        Z_grid[theory_grid - exp_grids['-2sigma'] < 0] = -2.5

        cp = ax.contour(X_grid, Y_grid, Z_grid, levels=[-2, -1, 0, 1, 2], linewidths=2, alpha=0, zorder=0)
        #ax.contourf(X_grid, Y_grid, Z_grid, levels=[-2, -1, 0, 1, 2], linewidths=2, alpha=1, zorder=0, cmap='GnBu')

        if draw_observed:
            Z_grid_obs = np.zeros(theory_grid.shape)
            Z_grid_obs[theory_grid - exp_grids['obs'] >= 0] = 0.5
            Z_grid_obs[theory_grid - exp_grids['obs'] < 0] = -0.5

            cp_obs = ax.contour(X_grid, Y_grid, Z_grid_obs, levels=[0], linewidths=2, alpha=0, zorder=0)

        # Plot 1 and 2 sigma bands
        self.plot_bands(ax, cp, labels, cp_obs if draw_observed else None)


    def draw(self, x:str="", y:str="", xlabel:str="", ylabel:str="", scale_factor=None, theory_grid=None, ylim=None, xlim=None,
             draw_observed:bool=True, observed_marker:Optional[str]='o', draw_sm_line:bool=False):
        
        ax = self.draw_frame()
        
        if self.data_sec is not None:
            self.draw_single_data(ax, df=self.data_sec, x=x, y=y,
                                  scale_factor=scale_factor,
                                  theory_grid=theory_grid,
                                  draw_observed=draw_observed,
                                  color_pallete=self.color_pallete_sec,
                                  labels=self.labels_sec,
                                  observed_marker=observed_marker, 
                                  sec=True)
            alpha = self.config['primary_alpha']
        else:
            alpha = 1.
        self.draw_single_data(ax, df=self.data, x=x, y=y,
                              scale_factor=scale_factor,
                              theory_grid=theory_grid,
                              draw_observed=draw_observed,
                              color_pallete=self.color_pallete,
                              labels=self.labels,
                              observed_marker=observed_marker, 
                              alpha=alpha)
        if self.highlight_data is not None:
            self.draw_highlight(ax, self.highlight_data)
            
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        
        if ylim is not None:
            ax.set_ylim(*ylim)
        if xlim is not None:
            ax.set_xlim(*xlim)

        if draw_sm_line:
            sm_line_styles = self.config['sm_line_styles']
            sm_values = self.config['sm_values']
            transform = create_transform(transform_y="axis", transform_x="data")
            ax.vlines(sm_values[0], ymin=0, ymax=1, zorder=2, transform=transform,
                      **sm_line_styles)
            transform = create_transform(transform_x="axis", transform_y="data")
            ax.hlines(sm_values[1], xmin=0, xmax=1, zorder=2, transform=transform,
                      **sm_line_styles)

        # border for the legend
        border_leg = patches.Rectangle((0, 0), 1, 1, facecolor = 'none', edgecolor = 'black', linewidth = 1)
        self.legend_data['1sigma']['handle'] = (self.legend_data['1sigma']['handle'], border_leg)
        self.legend_data['2sigma']['handle'] = (self.legend_data['2sigma']['handle'], border_leg)
        
        if self.data_sec is not None:
            self.legend_data_sec['1sigma']['handle'] = (self.legend_data_sec['1sigma']['handle'], border_leg)
            self.legend_data_sec['2sigma']['handle'] = (self.legend_data_sec['2sigma']['handle'], border_leg)
            
        handles, labels = self.get_legend_handles_labels()
        if self.data_sec is not None:
            handles_sec, labels_sec = self.get_legend_handles_labels(sec=True)
            handles = handles + handles_sec
            labels  = labels + labels_sec
        ax.legend(handles, labels, **self.styles['legend'])

        return ax




from ._version import __version__

import os
import pathlib
"""
import ROOT
ROOT.gROOT.SetBatch(True) 
ROOT.PyConfig.IgnoreCommandLineOptions = True
"""
module_path = pathlib.Path(__file__).parent.absolute()
macro_path = os.path.join(module_path, 'macros')
stylesheet_path = os.path.join(module_path, 'stylesheets')
resource_path = os.path.join(module_path, 'resources')

# ROOT.gInterpreter.AddIncludePath(os.path.join(macro_path, "macros"))

MAX_WORKERS = 8

from quickstats.utils.io import VerbosePrint
_PRINT_ = VerbosePrint("INFO")

corelib_loaded = False

from quickstats.core import *

root_version = get_root_version()


from typing import Optional, Union, List, Dict, Callable

import ROOT

from quickstats import PathManager
from quickstats.components import ROOTObject
from quickstats.interface.root import RooRealVar

class ComponentSource(ROOTObject):
    
    """

    class ModelType(DescriptiveEnum):
        SIGNAL     = (0, "Signal Model")
        BACKGROUND = (1, "Background Model")
        AUXILIARY  = (2, "Auxiliary Model")
    
                
    @property
    def model_type(self) -> ModelType:
        return self._model_type
    
    @model_type.setter
    def model_type(self, value:Union[str, ModelType]):
        if isinstance(value, ModelType):
            self._model_type = value
        else:
            self._model_type = ModelType.parse(value)
    """
    
    @property
    def observable(self) -> RooRealVar:
        return self._observable
    
    @observable.setter
    def observable(self, value:Union[RooRealVar, ROOT.RooRealVar, dict, str]):
        if value is None:
            self._observable = None
        elif isinstance(value, RooRealVar):
            self._observable = value
        else:
            self._observable = RooRealVar()
            self._observable.parse(value)
            
    @property
    def input_paths(self) -> PathManager:
        return self._input_paths
    
    @input_paths.setter
    def input_paths(self, value:Optional[Union[PathManager, str, List[str]]]=None):
        if value is None:
            self._input_paths = None
        if isinstance(value, PathManager):
            self._input_paths = value
        else:
            self._input_paths = PathManager(files=value)            
    
    def __init__(self, input_paths:Optional[Union[PathManager, str, List[str]]]=None,
                 observable:Optional[Union[RooRealVar, ROOT.RooRealVar, dict, str]]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        
        self.input_paths  = input_paths
        self.observable   = observable
    
    def get_resolved_input_paths(self, **file_kwargs):
        if self.input_paths is None:
            raise RuntimeError("input path(s) not set")
        full_paths = []
        for name in self.input_paths.files:
            path = self.input_paths.get_resolved_file(name, check_exist=True, **file_kwargs)
            full_paths.append(path)
        return full_paths
        
    def set_observable(self, observable:Union[RooRealVar, ROOT.RooRealVar]):
        self.observable = observable


from typing import Dict, List, Optional, Union

from .decorators import semistaticmethod
from .abstract_object import AbstractObject
from .configs import ConfigParser

class ConfigurableObject(AbstractObject):
    
    CONFIG_FORMAT = {
    }
    
    REQUIRED_CONFIG_COMPONENTS = []
    
    OPTIONAL_CONFIG_COMPONENTS = []
    
    def __init__(self, config_source:Optional[Union[Dict, str]]=None,
                 disable_config_message:bool=False,
                 verbosity:Optional[Union[int, str]]="INFO",
                 **kwargs) -> None:
        
        super().__init__(verbosity=verbosity)
        
        config_parser = ConfigParser(config_format=self.CONFIG_FORMAT,
                                     required_components=self.REQUIRED_CONFIG_COMPONENTS,
                                     optional_components=self.OPTIONAL_CONFIG_COMPONENTS,
                                     disable_message=disable_config_message)
        self.config_parser = config_parser
        if config_source is not None:
            self.load_config(config_source)
        else:
            self.config = None
        
    @semistaticmethod
    def explain_config(self, node_label:Optional[str]="",
                       required_only:bool=True,
                       indent:str="    ", bullet:str="=>",
                       linebreak:int=100) -> None:
        # call from class
        if not isinstance(self, ConfigurableObject):
            config_parser = ConfigParser(config_format=self.CONFIG_FORMAT,
                                         required_components=self.REQUIRED_CONFIG_COMPONENTS,
                                         optional_components=self.OPTIONAL_CONFIG_COMPONENTS)
            config_parser.explain(node_label, required_only=required_only,
                                  indent=indent, bullet=bullet, linebreak=linebreak)
        else:
            self.config_parser.explain(node_label, required_only=required_only,
                                       indent=indent, bullet=bullet, linebreak=linebreak)
        
    def load_config(self, config_source:Optional[Union[Dict, str]]) -> None:
        self.config = self.config_parser.parse_config(config_source)


from .rooproc_base_action import RooProcBaseAction
from .rooproc_rdf_action import RooProcRDFAction
from .rooproc_helper_action import RooProcHelperAction
from .rooproc_hybrid_action import RooProcHybridAction
from .rooproc_nested_action import RooProcNestedAction
from .rooproc_treename import RooProcTreeName
from .rooproc_declare import RooProcDeclare
from .rooproc_global_variables import RooProcGlobalVariables
from .rooproc_alias import RooProcAlias
from .rooproc_safe_alias import RooProcSafeAlias
from .rooproc_define import RooProcDefine
from .rooproc_safe_define import RooProcSafeDefine
from .rooproc_redefine import RooProcRedefine
from .rooproc_filter import RooProcFilter
from .rooproc_sum import RooProcSum
from .rooproc_max import RooProcMax
from .rooproc_min import RooProcMin
from .rooproc_mean import RooProcMean
from .rooproc_save import RooProcSave
from .rooproc_report import RooProcReport
from .rooproc_export import RooProcExport
from .rooproc_save_frame import RooProcSaveFrame
from .rooproc_load_frame import RooProcLoadFrame
from .rooproc_as_numpy import RooProcAsNumpy
from .rooproc_if_defined import RooProcIfDefined
from .rooproc_if_not_defined import RooProcIfNotDefined
from .auxiliary import *

ACTION_MAP = {
    "TREENAME": RooProcTreeName,
    "DECLARE": RooProcDeclare,
    "GLOBAL": RooProcGlobalVariables,
    "ALIAS": RooProcAlias,
    "SAFEALIAS": RooProcSafeAlias,
    "DEFINE": RooProcDefine,
    "SAFEDEFINE": RooProcSafeDefine,    
    "REDEFINE": RooProcRedefine,
    "FILTER": RooProcFilter,
    "GETSUM": RooProcSum,
    "GETMAX": RooProcMax,
    "GETMIN": RooProcMin,
    "GETMEAN": RooProcMean,
    "SAVE": RooProcSave,
    "REPORT": RooProcReport,    
    "EXPORT": RooProcExport,
    "SAVE_FRAME": RooProcSaveFrame,
    "LOAD_FRAME": RooProcLoadFrame,
    "AS_NUMPY": RooProcAsNumpy,
    "IFDEF": RooProcIfDefined,
    "IFNDEF": RooProcIfNotDefined    
}

def get_action(action_name:str):
    return ACTION_MAP.get(action_name, None)


from typing import Optional, Union, List, Dict, Callable

import ROOT

from quickstats import AbstractObject, PathManager
from quickstats.interface.root import RooRealVar
from .component_source import ComponentSource

class ModelSource(ComponentSource):
    
    def __init__(self, input_paths:Optional[Union[PathManager, str, List[str]]],
                 observable:Optional[RooRealVar]=None,
                 weight_name:Optional[str]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(input_paths=input_paths,
                         observable=observable,
                         verbosity=verbosity)
        self.weight_name = weight_name
        
    def create_model(self, workspace:ROOT.RooWorkspace):
        pass


from typing import Dict, Union, List, Optional

import numpy as np

from quickstats import semistaticmethod
from quickstats.interface.cppyy.vectorize import as_np_array

class RooArgSet:
    
    @staticmethod
    def from_list(args:List["ROOT.RooAbsArg"]):
        import ROOT
        return ROOT.RooArgSet(*args)
    
    @staticmethod
    def sort(argset:"ROOT.RooArgSet"):
        argset.sort()
        return argset


from quickstats import GeneralEnum, DescriptiveEnum

class WSVariables(GeneralEnum):
    VARIABLES                         = 0
    OBSERVABLES                       = 1
    POIS                              = 2
    GLOBAL_OBSERVABLES                = 3
    NUISANCE_PARAMETERS               = 4
    CONSTRAINED_NUISANCE_PARAMETERS   = 5
    UNCONSTRAINED_NUISANCE_PARAMETERS = 6
    CONSTRAINTS                       = 7
    AUXILIARY                         = 8
    CORE                              = 9
    NON_CONSTRAINT_VARIABLES          = 10
    
class WSArgument(DescriptiveEnum):
    WORKSPACE                         = (0, "Workspace", False)
    DATASET                           = (1, "Datasets", False)
    SNAPSHOT                          = (2, "Snapshots", False)
    PDF                               = (3, "PDFs", False)
    FUNCTION                          = (4, "Functions", False)
    VARIABLE                          = (5, "Workspace variables", True)
    OBSERVABLE                        = (6, "Observables", True)
    POI                               = (7, "Parameters of Interest (POI)", True)
    GLOBAL_OBSERVABLE                 = (8, "Global observables", True)
    NUISANCE_PARAMETER                = (9, "Nuisance parameters", True)
    CONSTRAINED_NUISANCE_PARAMETER    = (10, "Nuisance parameters with an associated constraint Pdf", True)
    UNCONSTRAINED_NUISANCE_PARAMETER  = (11, "Nuisance parameters without an associated constraint Pdf", True)
    CONSTRAINT                        = (12, "Tuple of constraint pdf and the associated nuisance parameter and global observable", False)
    AUXILIARY                         = (13, "Auxiliary variables (All variables except POIs, observables, global observables and "
                                             "nuisance parameters)", True)
    CORE                              = (14, "Nuisance parameters, global observables and POIs", True)
    MUTABLE                           = (15, "Mutable parameters (Nuisance parameters, global observables, POIs and floating auxiliary"
                                             " parameters)", True)
    
    def __new__(cls, value:int, description:str="", is_variable:bool=False):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        obj.is_variable = is_variable
        return obj

class SetValueMode(GeneralEnum):
    UNCHANGED = 0
    FIX       = 1
    FREE      = 2


from typing import List, Union, Optional, Dict
import os
import glob
import json
import shutil

import quickstats

def compile_macros(macros:Optional[Union[str, List[str]]]=None):
    """
    Compile ROOT macros
    
    Arguments:
        macros: (Optional) str or list of str
            If str, it is a string containing comma delimited list of macro names to be compiled.
            If list of str, it is the list of macro names to be compiled.
    """
    custom_list = True
    if macros is None:
        macros = get_all_macros()
        custom_list = False
    elif isinstance(macros, str):
        macros = macros.split(',')
    from quickstats.utils.root_utils import compile_macro
    for macro in macros:
        if ((macro == "FlexibleInterpVarMkII") and (quickstats.root_version >= (6, 26, 0))) and (not custom_list):
            quickstats._PRINT_.info("INFO: Skip compiling macro \"FlexibleInterpVarMkII\" which is "
                                    "deprecated since ROOT 6.26/00")
            continue
        compile_macro(macro)
        
def get_all_macros():
    """
    Get the list of macros names
    
    Note: Only macros ending in .cxx will be considered
    """
    macro_dir = quickstats.macro_path
    macro_subdirs = [path for path in glob.glob(os.path.join(macro_dir, "*")) if os.path.isdir(path)]
    macro_names = []
    for macro_subdir in macro_subdirs:
        macro_name = os.path.basename(macro_subdir)
        macro_path = os.path.join(macro_subdir, f"{macro_name}.cxx")
        if os.path.exists(macro_path):
            macro_names.append(macro_name)
    return macro_names

def set_verbosity(verbosity:Union[str, int]="INFO"):
    quickstats._PRINT_.verbosity = verbosity
    
def get_root_version():
    from quickstats.root_checker import ROOTChecker, ROOTVersion
    try:
        root_config_cmd = ROOTChecker.get_root_config_cmd()
        root_version = ROOTChecker.get_installed_root_version(root_config_cmd)
    except:
        root_version = ROOTVersion((0, 0, 0))
    return root_version

def get_workspace_extensions():
    extension_config = get_workspace_extension_config()
    extensions = extension_config['required']
    if not extension_config['strict']:
        # deprecated extension
        if (quickstats.root_version >= (6, 26, 0)) and 'FlexibleInterpVarMkII' in extensions:
            extensions.remove('FlexibleInterpVarMkII')
    return extensions

def get_workspace_extension_config():
    resource_path = quickstats.resource_path
    extension_config_file = os.path.join(resource_path, "workspace_extensions.json")
    if not os.path.exists(extension_config_file):
        extension_config_file = os.path.join(resource_path, "default_workspace_extensions.json")
    with open(extension_config_file, "r") as file:
        extension_config = json.load(file)
    return extension_config

def add_macro(path:str, name:Optional[str]=None, copy_files:bool=True,
              workspace_extension:bool=True, force:bool=False):
    if not os.path.isdir(path):
        raise ValueError("macro path must be a directory")
    if name is None:
        name = os.path.basename(os.path.abspath(path))
    source_path = os.path.join(path, f"{name}.cxx")
    if not os.path.exists(source_path):
        raise FileNotFoundError(f"macro path must contain a source file named {name}.cxx")
    macro_path = quickstats.macro_path
    if copy_files:
        dest_path = os.path.join(macro_path, name)
        if os.path.exists(dest_path):
            if force or input(f'WARNING: The macro already exists in {dest_path}, overwrite? [Y/N]') == "Y":
                shutil.rmtree(dest_path)
                shutil.copytree(path, dest_path)
                quickstats._PRINT_.info(f'INFO: Overwritten contents for the macro "{name}" from {path} to {dest_path}.')
            else:
                quickstats._PRINT_.info(f'INFO: Overwrite cancelled.')
        else:
            shutil.copytree(path, dest_path)
            quickstats._PRINT_.info(f'INFO: Copied contents for the macro "{name}" from {path} to {dest_path}.')
    if workspace_extension:
        resource_path = quickstats.resource_path
        extension_config_file = os.path.join(resource_path, "workspace_extensions.json")
        extension_config = get_workspace_extension_config()
        if name not in extension_config['required']:
            extension_config['required'].append(name)
            with open(extension_config_file, "w") as file:
                json.dump(extension_config, file, indent=2)
        quickstats._PRINT_.info(f'INFO: The macro "{name}" has been added to the workspace extension list.')
        
def remove_macro(name:str, remove_files:bool=True, force:bool=False):
    resource_path = quickstats.resource_path
    extension_config_file = os.path.join(resource_path, "workspace_extensions.json")
    extension_config = get_workspace_extension_config()
    if (name not in extension_config):
        quickstats._PRINT_.info(f'WARNING: Extension "{name}" not found in the workspace extension list. Skipped.')
    else:
        extension_config.remove(name)
        with open(extension_config_file, "w") as file:
            json.dump(extension_config, file, indent=2)
        quickstats._PRINT_.info(f'INFO: The extension "{name}" has been removed from the workspace extension list.')
    if remove_files:
        macro_path = quickstats.macro_path
        extension_path = os.path.abspath(os.path.join(macro_path, name))
        if force or input(f'WARNING: Attempting to remove files from {extension_path}, confirm? [Y/N]') == "Y":
            quickstats._PRINT_.info(f'INFO: Files for the macro "{name}" has been removed from {extension_path}.')
            shutil.rmtree(extension_path)
        else:
            quickstats._PRINT_.info(f'INFO: Macro files removal cancelled.')

def load_corelib():
    if not quickstats.corelib_loaded:
        from quickstats.utils.root_utils import load_macro
        load_macro("QuickStatsCore")
        quickstats.corelib_loaded = True
        
def load_extensions():
    extensions = get_workspace_extensions()
    extensions.extend(["QuickStatsCore", "AsymptoticCLsTool"])
    from quickstats.utils.root_utils import load_macro
    for extension in extensions:
        load_macro(extension)
        
def load_processor_methods():
    from quickstats.components.processors.builtin_methods import BUILTIN_METHODS
    from quickstats.utils.root_utils import declare_expression
    for name, definition in BUILTIN_METHODS.items():
        declare_expression(definition, name)    


from typing import Dict, Union, List, Optional

class RooMsgService:

    @staticmethod
    def remove_topics(input_arguments:bool=True, numeric_integration:bool=True, object_handling:bool=True):
        import ROOT
        if input_arguments:
            ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.InputArguments)
        if numeric_integration:
            ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.NumIntegration)
        if object_handling:
            ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.ObjectHandling)


from typing import Optional, Union, List, Dict

import ROOT

from quickstats import DescriptiveEnum, AbstractObject
from quickstats.maths.numerics import is_float

from .settings import (OBS_SET, POI_SET, NUIS_SET, GLOB_SET)

class CoreArgumentSets(AbstractObject):
    
    def __init__(self, verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)
        self.obs_set    = ROOT.RooArgSet()
        self.poi_set    = ROOT.RooArgSet()
        self.nuis_set   = ROOT.RooArgSet()
        self.glob_set   = ROOT.RooArgSet()
        self.constr_set = ROOT.RooArgSet()
        
    def get_argset_by_names(self, ws:ROOT.RooWorkspace, var_names:[str, List[str]], strict:bool=False):
        if isinstance(var_names, str):
            var_names = [var_names]
        argset = ROOT.RooArgSet()
        for var_name in var_names:
            var = ws.var(var_name)
            if not var:
                if strict:
                    raise RuntimeError(f'the variable "{var_name}" is not defined in the workspace "{ws.GetName()}"')
                continue
            argset.add(var)
        return argset
    
    def validate(self):
        if self.obs_set.size() == 0:
            raise RuntimeError('observable set not initialized')
        if self.poi_set.size() == 0:
            raise RuntimeError('POIs set not initialized')
   
    def define_argsets(self, ws:ROOT.RooWorkspace):
        self.validate()
        for argset, name in [(self.obs_set, OBS_SET), (self.poi_set, POI_SET),
                             (self.nuis_set, NUIS_SET), (self.glob_set, GLOB_SET)]:
            ws.defineSet(name, argset)
        self.stdout.debug(f'DEBUG: Defined argument set "{name}" for the workspace "{ws.GetName()}".')
        


from typing import Optional, Dict

from .rooproc_base_action import RooProcBaseAction

class RooProcHybridAction(RooProcBaseAction):
    
    def _execute(self, rdf:"ROOT.RDataFrame", processor:"quickstats.RooProcessor", **params):
        return rdf, processor
    
    def execute(self, rdf:"ROOT.RDataFrame", processor:"quickstats.RooProcessor", global_vars:Optional[Dict]=None):
        params = self.get_formatted_parameters(global_vars)
        rdf_next, processor_next = self._execute(rdf, processor, **params)
        self.executed = True
        return rdf_next, processor_next